"""Scaling solutions for production analysis from unconventional oil and gas wells."""

from __future__ import annotations

from importlib.metadata import version

__version__ = version("bluebonnet")

import bluebonnet.plotting
